@@ -11,6 +11,8 @@
 #include "xs_time.h"
 #include "xs_regex.h"
 #include "xs_match.h"
+#include "xs_unicode.h"
+#include "xs_random.h"
 
 #include "snac.h"
 
@@ -47,7 +49,7 @@ int srv_open(const char *basedir, int auto_upgrade)
     cfg_file = xs_fmt("%s/server.json", basedir);
 
     if ((f = fopen(cfg_file, "r")) == NULL)
-        error = xs_fmt("ERROR: cannot opening '%s'", cfg_file);
+        error = xs_fmt("ERROR: cannot open '%s'", cfg_file);
     else {
         /* read full config file */
         srv_config = xs_json_load(f);
@@ -155,6 +157,22 @@ int srv_open(const char *basedir, int auto_upgrade)
     /* read (and drop) emojis.json, possibly creating it */
     xs_free(emojis());
 
+    /* if style.css does not exist, create it */
+    xs *css_fn = xs_fmt("%s/style.css", srv_basedir);
+
+    if (mtime(css_fn) == 0) {
+        srv_log(xs_fmt("Writing style.css"));
+        write_default_css();
+    }
+
+    /* create the proxy token seed */
+    {
+        char rnd[16];
+        xs_rnd_buf(rnd, sizeof(rnd));
+
+        srv_proxy_token_seed = xs_hex_enc(rnd, sizeof(rnd));
+    }
+
     return ret;
 }
 
@@ -328,6 +346,7 @@ int user_persist(snac *snac, int publish)
         rename(bfn, fn);
 
     history_del(snac, "timeline.html_");
+    timeline_touch(snac);
 
     if (publish) {
         xs *a_msg = msg_actor(snac);
@@ -780,7 +799,7 @@ int _object_add(const char *id, const xs_dict *obj, int ow)
         fclose(f);
 
         /* does this object has a parent? */
-        const char *in_reply_to = xs_dict_get(obj, "inReplyTo");
+        const char *in_reply_to = get_in_reply_to(obj);
 
         if (!xs_is_null(in_reply_to) && *in_reply_to) {
             /* update the children index of the parent */
@@ -1012,13 +1031,31 @@ int object_unadmire(const char *id, const char *actor, int like)
 }
 
 
-int _object_user_cache(snac *snac, const char *id, const char *cachedir, int del)
+xs_str *object_user_cache_fn_by_md5(snac *user, const char *md5, const char *cachedir)
+{
+    return xs_fmt("%s/%s/%s.json", user->basedir, cachedir, md5);
+}
+
+
+xs_str *object_user_cache_fn(snac *user, const char *id, const char *cachedir)
+{
+    xs *md5 = xs_md5_hex(id, strlen(id));
+    return object_user_cache_fn_by_md5(user, md5, cachedir);
+}
+
+
+xs_str *object_user_cache_index_fn(snac *user, const char *cachedir)
+{
+    return xs_fmt("%s/%s.idx", user->basedir, cachedir);
+}
+
+
+int _object_user_cache(snac *user, const char *id, const char *cachedir, int del)
 /* adds or deletes from a user cache */
 {
     xs *ofn = _object_fn(id);
-    xs *l   = xs_split(ofn, "/");
-    xs *cfn = xs_fmt("%s/%s/%s", snac->basedir, cachedir, xs_list_get(l, -1));
-    xs *idx = xs_fmt("%s/%s.idx", snac->basedir, cachedir);
+    xs *cfn = object_user_cache_fn(user, id, cachedir);
+    xs *idx = object_user_cache_index_fn(user, cachedir);
     int ret;
 
     if (del) {
@@ -1026,6 +1063,10 @@ int _object_user_cache(snac *snac, const char *id, const char *cachedir, int del
         index_del(idx, id);
     }
     else {
+        /* create the subfolder, if it does not exist */
+        xs *dir = xs_fmt("%s/%s/", user->basedir, cachedir);
+        mkdirx(dir);
+
         if ((ret = link(ofn, cfn)) != -1)
             index_add(idx, id);
     }
@@ -1034,34 +1075,40 @@ int _object_user_cache(snac *snac, const char *id, const char *cachedir, int del
 }
 
 
-int object_user_cache_add(snac *snac, const char *id, const char *cachedir)
+int object_user_cache_add(snac *user, const char *id, const char *cachedir)
 /* caches an object into a user cache */
 {
-    return _object_user_cache(snac, id, cachedir, 0);
+    return _object_user_cache(user, id, cachedir, 0);
 }
 
 
-int object_user_cache_del(snac *snac, const char *id, const char *cachedir)
+int object_user_cache_del(snac *user, const char *id, const char *cachedir)
 /* deletes an object from a user cache */
 {
-    return _object_user_cache(snac, id, cachedir, 1);
+    return _object_user_cache(user, id, cachedir, 1);
 }
 
 
-int object_user_cache_in(snac *snac, const char *id, const char *cachedir)
+int object_user_cache_in(snac *user, const char *id, const char *cachedir)
 /* checks if an object is stored in a cache */
 {
-    xs *md5 = xs_md5_hex(id, strlen(id));
-    xs *cfn = xs_fmt("%s/%s/%s.json", snac->basedir, cachedir, md5);
-
+    xs *cfn = object_user_cache_fn(user, id, cachedir);
     return !!(mtime(cfn) != 0.0);
 }
 
 
-xs_list *object_user_cache_list(snac *snac, const char *cachedir, int max, int inv)
+int object_user_cache_in_by_md5(snac *user, const char *md5, const char *cachedir)
+/* checks if an object is stored in a cache */
+{
+    xs *cfn = object_user_cache_fn_by_md5(user, md5, cachedir);
+    return !!(mtime(cfn) != 0.0);
+}
+
+
+xs_list *object_user_cache_list(snac *user, const char *cachedir, int max, int inv)
 /* returns the objects in a cache as a list */
 {
-    xs *idx = xs_fmt("%s/%s.idx", snac->basedir, cachedir);
+    xs *idx = xs_fmt("%s/%s.idx", user->basedir, cachedir);
     return inv ? index_list_desc(idx, 0, max) : index_list(idx, max);
 }
 
@@ -1203,6 +1250,9 @@ int timeline_del(snac *snac, const char *id)
     object_user_cache_del(snac, id, "public");
     object_user_cache_del(snac, id, "private");
 
+    unpin(snac, id);
+    unbookmark(snac, id);
+
     /* try to delete the object if it's not used elsewhere */
     return object_del_if_unref(id);
 }
@@ -1300,10 +1350,17 @@ xs_list *timeline_top_level(snac *snac, const xs_list *list)
 }
 
 
-xs_list *timeline_simple_list(snac *snac, const char *idx_name, int skip, int show)
+xs_str *user_index_fn(snac *user, const char *idx_name)
+/* returns the filename of a user index */
+{
+    return xs_fmt("%s/%s.idx", user->basedir, idx_name);
+}
+
+
+xs_list *timeline_simple_list(snac *user, const char *idx_name, int skip, int show)
 /* returns a timeline (with all entries) */
 {
-    xs *idx = xs_fmt("%s/%s.idx", snac->basedir, idx_name);
+    xs *idx = user_index_fn(user, idx_name);
 
     return index_list_desc(idx, skip, show);
 }
@@ -1327,10 +1384,16 @@ xs_list *timeline_list(snac *snac, const char *idx_name, int skip, int show)
 }
 
 
+xs_str *instance_index_fn(void)
+{
+    return xs_fmt("%s/public.idx", srv_basedir);
+}
+
+
 xs_list *timeline_instance_list(int skip, int show)
 /* returns the timeline for the full instance */
 {
-    xs *idx = xs_fmt("%s/public.idx", srv_basedir);
+    xs *idx = instance_index_fn();
 
     return index_list_desc(idx, skip, show);
 }
@@ -1521,27 +1584,80 @@ int is_muted(snac *snac, const char *actor)
 }
 
 
-/** pinning **/
-
-xs_str *_pinned_fn(snac *user, const char *id)
+xs_list *muted_list(snac *user)
+/* returns the list (actor URLs) of the muted morons */
 {
-    xs *md5 = xs_md5_hex(id, strlen(id));
-    return xs_fmt("%s/pinned/%s.json", user->basedir, md5);
+    xs_list *l = xs_list_new();
+    xs *spec = xs_fmt("%s/muted/" "*", user->basedir);
+    xs *files = xs_glob(spec, 0, 0);
+    const char *fn;
+
+    xs_list_foreach(files, fn) {
+        FILE *f;
+
+        if ((f = fopen(fn, "r")) != NULL) {
+            xs *actor = xs_strip_i(xs_readline(f));
+            fclose(f);
+
+            l = xs_list_append(l, actor);
+        }
+    }
+
+    return l;
 }
 
 
+/** bookmarking **/
+
+int is_bookmarked(snac *user, const char *id)
+/* returns true if this note is bookmarked */
+{
+    return object_user_cache_in(user, id, "bookmark");
+}
+
+
+int bookmark(snac *user, const char *id)
+/* bookmarks a post */
+{
+    if (is_bookmarked(user, id))
+        return -3;
+
+    return object_user_cache_add(user, id, "bookmark");
+}
+
+
+int unbookmark(snac *user, const char *id)
+/* unbookmarks a post */
+{
+    return object_user_cache_del(user, id, "bookmark");
+}
+
+
+xs_list *bookmark_list(snac *user)
+/* return the lists of bookmarked posts */
+{
+    return object_user_cache_list(user, "bookmark", XS_ALL, 1);
+}
+
+
+xs_str *bookmark_index_fn(snac *user)
+{
+    return object_user_cache_index_fn(user, "bookmark");
+}
+
+
+/** pinning **/
+
 int is_pinned(snac *user, const char *id)
 /* returns true if this note is pinned */
 {
-    xs *fn = _pinned_fn(user, id);
-    return !!(mtime(fn) != 0.0);
+    return object_user_cache_in(user, id, "pinned");
 }
 
 
 int is_pinned_by_md5(snac *user, const char *md5)
 {
-    xs *fn = xs_fmt("%s/pinned/%s.json", user->basedir, md5);
-    return !!(mtime(fn) != 0.0);
+    return object_user_cache_in_by_md5(user, md5, "pinned");
 }
 
 
@@ -1553,13 +1669,8 @@ int pin(snac *user, const char *id)
     if (xs_startswith(id, user->actor)) {
         if (is_pinned(user, id))
             ret = -3;
-        else {
-            /* create the subfolder, if it does not exist */
-            xs *fn = xs_fmt("%s/pinned/", user->basedir);
-            mkdirx(fn);
-
+        else
             ret = object_user_cache_add(user, id, "pinned");
-        }
     }
 
     return ret;
@@ -1569,15 +1680,7 @@ int pin(snac *user, const char *id)
 int unpin(snac *user, const char *id)
 /* unpin a message */
 {
-    int ret = object_user_cache_del(user, id, "pinned");
-
-    if (ret != -1) {
-        /* delete from the index */
-        xs *idx = xs_fmt("%s/pinned.idx", user->basedir);
-        index_del(idx, id);
-    }
-
-    return ret;
+    return object_user_cache_del(user, id, "pinned");
 }
 
 
@@ -1588,6 +1691,45 @@ xs_list *pinned_list(snac *user)
 }
 
 
+/** drafts **/
+
+int is_draft(snac *user, const char *id)
+/* returns true if this note is a draft */
+{
+    return object_user_cache_in(user, id, "draft");
+}
+
+
+void draft_del(snac *user, const char *id)
+/* delete a message from the draft cache */
+{
+    object_user_cache_del(user, id, "draft");
+}
+
+
+void draft_add(snac *user, const char *id, const xs_dict *msg)
+/* store the message as a draft */
+{
+    /* delete from the index, in case it was already there */
+    draft_del(user, id);
+
+    /* overwrite object */
+    object_add_ow(id, msg);
+
+    /* [re]add to the index */
+    object_user_cache_add(user, id, "draft");
+}
+
+
+xs_list *draft_list(snac *user)
+/* return the lists of drafts */
+{
+    return object_user_cache_list(user, "draft", XS_ALL, 1);
+}
+
+
+/** hiding **/
+
 xs_str *_hidden_fn(snac *snac, const char *id)
 {
     xs *md5 = xs_md5_hex(id, strlen(id));
@@ -1761,6 +1903,7 @@ void tag_index(const char *id, const xs_dict *obj)
 /* update the tag indexes for this object */
 {
     const xs_list *tags = xs_dict_get(obj, "tag");
+    xs *md5_id = xs_md5_hex(id, strlen(id));
 
     if (is_msg_public(obj) && xs_type(tags) == XSTYPE_LIST && xs_list_len(tags) > 0) {
         xs *g_tag_dir = xs_fmt("%s/tag", srv_basedir);
@@ -1787,7 +1930,9 @@ void tag_index(const char *id, const xs_dict *obj)
                 mkdirx(tag_dir);
 
                 xs *g_tag_idx = xs_fmt("%s/%s.idx", tag_dir, md5_tag);
-                index_add(g_tag_idx, id);
+
+                if (!index_in_md5(g_tag_idx, md5_id))
+                    index_add_md5(g_tag_idx, md5_id);
 
                 FILE *f;
                 xs *g_tag_name = xs_replace(g_tag_idx, ".idx", ".tag");
@@ -1803,15 +1948,22 @@ void tag_index(const char *id, const xs_dict *obj)
 }
 
 
-xs_list *tag_search(const char *tag, int skip, int show)
-/* returns the list of posts tagged with tag */
+xs_str *tag_fn(const char *tag)
 {
     if (*tag == '#')
         tag++;
 
     xs *lw_tag = xs_tolower_i(xs_dup(tag));
     xs *md5    = xs_md5_hex(lw_tag, strlen(lw_tag));
-    xs *idx    = xs_fmt("%s/tag/%c%c/%s.idx", srv_basedir, md5[0], md5[1], md5);
+
+    return xs_fmt("%s/tag/%c%c/%s.idx", srv_basedir, md5[0], md5[1], md5);
+}
+
+
+xs_list *tag_search(const char *tag, int skip, int show)
+/* returns the list of posts tagged with tag */
+{
+    xs *idx = tag_fn(tag);
 
     return index_list_desc(idx, skip, show);
 }
@@ -1864,6 +2016,9 @@ xs_val *list_maint(snac *user, const char *list, int op)
             while (xs_list_next(lol, &v, &c)) {
                 if (strcmp(xs_list_get(v, 1), list) == 0) {
                     add = 0;
+
+                    l = xs_dup(xs_list_get(v, 0));
+
                     break;
                 }
             }
@@ -1924,6 +2079,12 @@ xs_val *list_maint(snac *user, const char *list, int op)
 }
 
 
+xs_str *list_timeline_fn(snac *user, const char *list)
+{
+    return xs_fmt("%s/list/%s.idx", user->basedir, list);
+}
+
+
 xs_list *list_timeline(snac *user, const char *list, int skip, int show)
 /* returns the timeline of a list */
 {
@@ -1932,10 +2093,12 @@ xs_list *list_timeline(snac *user, const char *list, int skip, int show)
     if (!xs_is_hex(list))
         return NULL;
 
-    xs *fn = xs_fmt("%s/list/%s.idx", user->basedir, list);
+    xs *fn = list_timeline_fn(user, list);
 
     if (mtime(fn) > 0.0)
         l = index_list_desc(fn, skip, show);
+    else
+        l = xs_list_new();
 
     return l;
 }
@@ -2388,7 +2551,7 @@ xs_list *content_search(snac *user, const char *regex,
     if (regex == NULL || *regex == '\0')
         return xs_list_new();
 
-    xs *i_regex = xs_tolower_i(xs_dup(regex));
+    xs *i_regex = xs_utf8_to_lower(regex);
 
     xs_set seen;
 
@@ -2460,18 +2623,36 @@ xs_list *content_search(snac *user, const char *regex,
         if (id == NULL || is_hidden(user, id))
             continue;
 
+        xs *c = xs_str_new(NULL);
         const char *content = xs_dict_get(post, "content");
+        const char *name    = xs_dict_get(post, "name");
 
-        if (xs_is_null(content))
-            continue;
+        if (!xs_is_null(content))
+            c = xs_str_cat(c, content);
+        if (!xs_is_null(name))
+            c = xs_str_cat(c, " ", name);
+
+        /* add alt-texts from attachments */
+        const xs_list *atts = xs_dict_get(post, "attachment");
+        int tc = 0;
+        const xs_dict *att;
+
+        while (xs_list_next(atts, &att, &tc)) {
+            const char *name = xs_dict_get(att, "name");
+
+            if (name != NULL)
+                c = xs_str_cat(c, " ", name);
+        }
 
         /* strip HTML */
-        xs *c = xs_regex_replace(content, "<[^>]+>", " ");
+        c = xs_regex_replace_i(c, "<[^>]+>", " ");
         c = xs_regex_replace_i(c, " {2,}", " ");
-        c = xs_tolower_i(c);
+
+        /* convert to lowercase */
+        xs *lc = xs_utf8_to_lower(c);
 
         /* apply regex */
-        if (xs_regex_match(c, i_regex)) {
+        if (xs_regex_match(lc, i_regex)) {
             if (xs_set_add(&seen, md5) == 1)
             show--;
         }
@@ -2785,7 +2966,7 @@ void enqueue_output_by_actor(snac *snac, const xs_dict *msg,
                             const xs_str *actor, int retries)
 /* enqueues an output message for an actor */
 {
-    xs *inbox = get_actor_inbox(actor);
+    xs *inbox = get_actor_inbox(actor, 1);
 
     if (!xs_is_null(inbox))
         enqueue_output(snac, msg, inbox, retries, 0);
@@ -3213,7 +3394,8 @@ void purge_user(snac *snac)
 
     _purge_user_subdir(snac, "public",  pub_days);
 
-    const char *idxs[] = { "followers.idx", "private.idx", "public.idx", "pinned.idx", NULL };
+    const char *idxs[] = { "followers.idx", "private.idx", "public.idx",
+                           "pinned.idx", "bookmark.idx", "draft.idx", NULL };
 
     for (n = 0; idxs[n]; n++) {
         xs *idx = xs_fmt("%s/%s", snac->basedir, idxs[n]);
@@ -3477,3 +3659,29 @@ t_announcement *announcement(const double after)
 
     return NULL;
 }
+
+
+xs_str *make_url(const char *href, const char *proxy, int by_token)
+/* makes an URL, possibly including proxying */
+{
+    xs_str *url = NULL;
+
+    if (proxy && !xs_startswith(href, srv_baseurl)) {
+        xs *p = NULL;
+
+        if (by_token) {
+            xs *tks = xs_fmt("%s:%s", srv_proxy_token_seed, proxy);
+            xs *tk = xs_md5_hex(tks, strlen(tks));
+
+            p = xs_fmt("%s/y/%s/", proxy, tk);
+        }
+        else
+            p = xs_fmt("%s/x/", proxy);
+
+        url = xs_replace(href, "https:/" "/", p);
+    }
+    else
+        url = xs_dup(href);
+
+    return url;
+}
